/* TPUset.c: TPU initialisation program for Kanis VTSplus QSPI functions
 *
 * Copyright (C) 2005  Sieghard Schicktanz <S.Schicktanz@kanis.de>,
 *                                                                      */
/* This program is free software; you can redistribute it and/or        */
/* modify it under the terms of the GNU Library General Public License  */
/* as published by the Free Software Foundation; either version 2       */
/* of the License, or (at your option) any later version.               */
/*                                                                      */
/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU Library General Public License for more details.                 */
/*
   Functions:

   Syntax:
     TPUset [control file]
   where
     control file: file containing commands and data

   Control file syntax:
     Lines containing TPU commands or Parameter RAM data like
        "MCR" [data]
        [TPU|PRAM] [channel] {[data]}*
     or comments like
        #[comment text]
	and
        ![comment text]
	    '!'-comments are displayed on execution
     or blank lines.
   where
     MCR:  specifies TPU master control register setup data
             Infulences ALL channels at once!
     TPU:  specifies setup data for TPU registers of channel "channel",
           consisting of exactly 8 values specifying TPU register data
	   according to the following list:
             function code, host sequence, host service, priority,
             interrupt enable, stop on close, set mask, get mask
     PRAM: specifies setup data for Parameter RAM words of channel "channel",
           up to 8 words according to value of "set mask"
     channel: [number] specifies the TPU channel the operation applies to
     data: [base specifier][value]
     base specifier: [$|#|o|%|'] - $: hex, #: decimal, o: octal, %: binary, ': character, default: hex
     comment text: unevaluated text till the end of the line
*/

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <asm/tpu_ioctl.h>


#define TPU_CHANNELS	16

static char tpu_dev [] = "/dev/tpu0";

unsigned short Parameters [10];


unsigned int scanvalue (char * number)
{
    if (!number) return 0;

    switch (tolower (number [0])) {
	case '$':
	    return strtoul (&(number [1]), NULL, 16);
	case '#':
	    return strtoul (&(number [1]), NULL, 10);
	case 'o':
	    return strtoul (&(number [1]), NULL, 8);
	case '%':
	    return strtoul (&(number [1]), NULL, 2);
	case '\'':
	    return number [1];
	default:
	    return strtoul (number, NULL, 0);
    }
    return 0;
}


void TPUio (FILE * cmds)
{
    char inpline [128], * token;
    unsigned int tpu = 0, channel = TPU_CHANNELS, i;

    while (fgets (inpline, sizeof (inpline), cmds)) {
	switch (tolower (inpline [0])) {
	    case '!':  /* comment line */
		printf (inpline);  /* show for reference */
	    case '#':  /* comment line */
	    case ' ':  /* empty line */
	    case '\n': /* empty line */
	    case '\t': /* empty line */
	    case '\0': /* empty line */
		break;
	    default:
		token= strtok (inpline, " \t\n");                // [TPU|PRAM]

		if (!strcasecmp ("MCR", token)) {
		    if ((token= strtok (NULL, " \t\n"))) {       // [data]
			if (!tpu) {                              // must have device open!
			    if (!(tpu= open (tpu_dev, O_RDWR))) {
				printf ("Cannot open TPU device #%c\n", tpu_dev [8]);
				channel= TPU_CHANNELS;
				break;
			    }
			    ioctl (tpu, TPUIOCS_SETMCR, scanvalue (token));
			    close (tpu);
			}
		    }
		}
		else if (!strcasecmp ("TPU", token)) {
		    if ((token= strtok (NULL, " \t\n"))) {       // [channel]
			i= channel; channel= scanvalue (token);
			tpu_dev [8]= (channel < 10)? '0'+ channel: 'a'+ (channel- 10);

			if (channel != i) {
			    if (i < TPU_CHANNELS) close (tpu);
			    if (!(tpu= open (tpu_dev, O_RDWR))) {
				printf ("Cannot open TPU device #%x\n", channel);
				channel= TPU_CHANNELS;
				break;
			    }
			}
			i= 0; memset (&Parameters, 0, sizeof (Parameters));
			while ((token= strtok (NULL, " \t\n")) && (i < sizeof (Parameters))) {
			    Parameters [i++]= scanvalue (token);
			}
			if (i) {
			    tpu_device TPU_default;

			    TPU_default.channel_func=  Parameters [0];   /* Channel function code */
			    TPU_default.host_sequence= Parameters [1];   /* Host sequence code */
			    TPU_default.host_service=  Parameters [2];   /* Host service code */
			    TPU_default.priority=      Parameters [3];   /* Channel priority */
			    TPU_default.interrupt=     Parameters [4];   /* Channel interrupt used? */
			    TPU_default.stop_on_close= Parameters [5];   /* Channel stops on close */
			    TPU_default.overwrite_old= Parameters [6];   /* Channel stops on close */
			    TPU_default.set_mask=      Parameters [7];   /* Channel data bit mask for parameter RAM setup */
			    TPU_default.get_mask=      Parameters [8];   /* Channel data bit mask for reading from parameter RAM */
			    TPU_default.clr_mask=      Parameters [9];   /* Channel data bit mask for clearing parameter RAM */

			    ioctl (tpu, TPUIOCS_SETPAR, &TPU_default);
			}
		    }
		}
		else if (!strcasecmp ("PRAM", token)) {
		    if ((token= strtok (NULL, " \t\n"))) {       // [channel]
			i= channel; channel= scanvalue (token);
			tpu_dev [8]= (channel < 10)? '0'+ channel: 'a'+ (channel- 10);

			if (channel != i) {
			    if (i < TPU_CHANNELS) close (tpu);
			    if (!(tpu= open (tpu_dev, O_RDWR))) {
				printf ("Cannot open TPU device #%x\n", channel);
				channel= TPU_CHANNELS;
				break;
			    }
			}
			i= 0;
			while ((token= strtok (NULL, " \t\n")) && (i < sizeof (Parameters))) {
			    Parameters [i++]= scanvalue (token);
			}
			write (tpu, Parameters, 2* i);
		    }
		}
		else fprintf (stderr, "Invalid command: %s", inpline);
	}
    }
    if (tpu) close (tpu);
}


int usage (void)
{
    puts ("\nSyntax:");
    puts ("\tTPUset [control file]");
    puts ("where");
    puts ("\tcontrol file: file containing I/O operation commands and data");
    puts ("");
    return 0;
}


int main (int argc, char *argv[], char *env[])
{
    FILE * cmds;
    int uid;

    puts ("Kanis VTSplus TPU setup program V.0.0 SchS/Kanis GmbH 022005");

    if (argc <= 1) return usage ();

    if ((uid= getuid ())) {    // are we root already
	printf ("You are not root! UID is %d, trying to set UID 0...", uid);

	if (seteuid (0))      // couldn't become root...
	    puts ("Cannot acquire root privileges!"); return 1;
    }
    cmds= fopen (argv [1], "r");
    if (cmds == NULL) {
	puts ("Cannot open control file!"); return 1;
    }
    TPUio (cmds);

    fclose (cmds);
    return 0;
}
