/************************************************************************/
/*                                                                      */
/*  cpu32tpu.c - TPU driver for MC68332                                 */
/*                                                                      */
/*  (C) Copyright 2004, Sieghard Schicktanz (S.Schicktanz@Kanis.de)     */
/*                                                                      */
/* This program is free software; you can redistribute it and/or        */
/* modify it under the terms of the GNU Library General Public License  */
/* as published by the Free Software Foundation; either version 2       */
/* of the License, or (at your option) any later version.               */
/*                                                                      */
/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU Library General Public License for more details.                 */
/*                                                                      */
/************************************************************************/

/*************************************************************************

TPU Register Assignments:
(CAUTION: assignment sequence from high to low!)

Channel	CFSR		Host Sequence	Host Service	Priority	Parameter RAM
 0	0xfffe12:3:0	0xfffe16:1:0	0xfffe1a:1:0	0xfffe1e:1:0	0xffff00-0xffff0c
 1	0xfffe12:7:4	0xfffe16:3:2	0xfffe1a:3:2	0xfffe1e:3:2	0xffff10-0xffff1c
 2	0xfffe12:11:8	0xfffe16:5:4	0xfffe1a:5:4	0xfffe1e:5:4	0xffff20-0xffff2c
 3	0xfffe12:15:12	0xfffe16:7:6	0xfffe1a:7.6	0xfffe1e:7:6	0xffff30-0xffff3c
 4	0xfffe10:3:0	0xfffe16:9:8	0xfffe1a:9:8	0xfffe1e:9:8	0xffff40-0xffff4c
 5	0xfffe10:7:4	0xfffe16:11:10	0xfffe1a:11:10	0xfffe1e:11:10	0xffff50-0xffff5c
 6	0xfffe10:11:8	0xfffe16:13:12	0xfffe1a:13:12	0xfffe1e:13:12	0xffff60-0xffff6c
 7	0xfffe10:15:12	0xfffe16:15:14	0xfffe1a:15:14	0xfffe1e:15:14	0xffff70-0xffff7c
 8	0xfffe0e:3:0	0xfffe14:1:0	0xfffe18:1:0	0xfffe1c:1:0	0xffff80-0xffff8c
 9	0xfffe0e:7:4	0xfffe14:3:2	0xfffe18:3:2	0xfffe1c:3:2	0xffff90-0xffff9c
10	0xfffe0e:11:8	0xfffe14:5:4	0xfffe18:5:4	0xfffe1c:5:4	0xffffa0-0xffffac
11	0xfffe0e:15:12	0xfffe14:7:6	0xfffe18:7:6	0xfffe1c:7:6	0xffffb0-0xffffbc
12	0xfffe0c:3:0	0xfffe14:9:8	0xfffe18:9:8	0xfffe1c:9:8	0xffffc0-0xffffcc
13	0xfffe0c:7:4	0xfffe14:11:10	0xfffe18:11:10	0xfffe1c:11:10	0xffffd0-0xffffdc
14	0xfffe0c:11:8	0xfffe14:13:12	0xfffe18:13:12	0xfffe1c:13:12	0xffffe0-0xffffee
15	0xfffe0c:15:12	0xfffe14:15:14	0xfffe18:15:14	0xfffe1c:15:14	0xfffff0-0xfffffe


Table A-1 Mask Set A Time Function Encodings

Function		Function	Host Service Request Code	Host Sequence
Name			Code						Code

PPWA			$F		0 = None			0 = 24-Bit Period
Period/Pulse-Width			1 = (Not Implemented)		1 = 16-Bit Period + Link
Accumulator				2 = Initialization		2 = 24-Bit Pulse Width
					3 = (Not Implemented)		3 = 16-Bit Pulse Width + Link

OC			$E		0 = None			0 = Execute All Functions
Output Compare				1 = Host-Initiated Pulse Mode	1 = Execute All Functions
					2 = (Not Implemented)		2 = Only Update TCRn Parameters
					3 = Continuous Pulse Mode	3 = Only Update TCRn Parameters

SM			$D		0 = None			(None Implemented)
Stepper Motor				1 = None
					2 = Initialization
					3 = Step Request

PSP			$C		0 = None			0 = Pulse Width Set by Angle
Position-Synchronized			1 = Immediate Update Request	1 = Pulse Width Set by Time
Pulse Generator				2 = Initialization		2 = Pulse Width Set by Angle
					3 = Force Change		3 = Pulse Width Set by Time

PMA/PMM			$B		0 = None			0 = PMA Bank Mode
Period Measurement			1 = Initialization		1 = PMA Count Mode
with Additional/Missing			2 = (Not Implemented)		2 = PMM Bank Mode
Transition Detect			3 = (Not Implemented)		3 = PMM Count Mode

ITC			$A		0 = None			0 = No Link, Single Mode
Input Capture/				1 = Initialization		1 = No Link, Continuous Mode
Input Transition Counter		2 = (Not Implemented)		2 = Link, Single Mode
					3 = (Not Implemented)		3 = Link, Continuous Mode

PWM			$9		0 = None			(None Implemented)
Pulse-Width Modulation			1 = Immediate Update Request
					2 = Initialization
					3 = (Not Implemented)

DIO			$8		0 = None			0 = Trans Mode  Record Pin on
Discrete Input/Output			1 = Force Output High		Transition
					2 = Force Output Low		0 = Trans Mode  Record Pin on
					3 = Init, Input Specified	Transition
					3 = Init, Periodic Input	0 = Trans Mode  Record Pin on
					3 = Update Pin Status Parameter	Transition
									1 = Match Mode  Record Pin at
									    MATCH_RATE
									2 = Record Pin State on HSR 11

SPWM			$7		0 = None			0 = Mode 0
Synchronized Pulse-			1 = (Not Implemented)		1 = Mode 1
Width Modulation			2 = Initialization		2 = Mode 2
					3 = Immediate Update Request	3 = (Not Implemented)

QDEC			$6		0x = No Action			x0 = Primary Channel
Quadrature Decode			10 = Read TCR1			x1 = Secondary Channel
					11 = Initialize

Table A-2 Mask Set G Time Function Encodings

Function		Function	Host Service Request Code	Host Sequence
Name			Code						Code

PTA			$F		0 = No host service		0 = High time accumulate
Programmable Time			1 = No effect			1 = Low time accumulate
Accumulator				2 = No effect			2 = Period accumulate  rising
					3 = Initialize function		3 = Period accumulate  falling

Queued Output		$E		0 = No Host Service		0 = Single-shot mode
Match (QOM)				1 = Initialize, No Pin Change	1 = Loop Mode
					2 = Initialize, Pin Low		2 = Continuous Mode
					3 = Initialize, Pin High	3 = Continuous Mode

TSM			$D		0 = No Host Service		0 = Rotate PIN_SEQUENCE once
Table Stepper Motor			1 = Initialize, Pin Low		    between steps, local mode
					2 = Initialize, Pin High	    acceleration table
					3 = Move Request (Master Only)	1 = Rotate PIN_SEQUENCE once
									    between steps, split mode
									    acceleration table
									2 = Rotate PIN_SEQUENCE twice
									    between steps, local mode
									    acceleration table
									3 = Rotate PIN_SEQUENCE twice
									    between steps, split mode
									    acceleration table

FQM			$C		0 = No Host Service		0 = Begin with Falling Edge Single-
Frequency Measurement			1 = Undefined			    Shot Mode
					2 = Initialize			1 = Begin with Falling Edge 
					3 = Undefined			    Continuous Mode
									2 = Begin with Rising Edge  Single-
									    Shot Mode
									3 = Begin with Rising Edge 
									    Continuous Mode

UART			$B		0 = No Host Service		0 = No Parity
Asynchronous				1 = Not used			1 = No Parity
Receiver/Transmitter			2 = Receive			2 = Even Parity
					3 = Transmit			3 = Odd Parity

NITC			$A		0 = No Host Service		0 = Single Shot, No Links
New Input Transition			1 = Initialize TCR Mode		1 = Continual, No Links
Counter					2 = Initialize Parameter Mode	2 = Single Shot, Links
					3 = Not Used			3 = Continual, Links

COMM			$9		0 = No host service request	0 = Sensorless match update mode
Multiphase Motor			1 = Not used			1 = Sensorless match update mode
Commutation				2 = Initialize or force state	2 = Sensorless link update mode
					3 = Initialize or force immed	3 = Sensored mode
					    state test

HALLD			$8		0 = No host service		0 = Channel A
					1 = Not used			1 = Channel B
					2 = Initialize two channel mode	2 = Channel B
					3 = Initialize three channel	3 = Channel C (3-channel mode
					    mode			    only)

MCPWM			$7		0 = No Host Service		0 = Edge-Aligned Mode
Multichannel PWM			1 = Init as Slave (Inverted)	1 = Slave A Type CA Mode
					2 = Init as Slave (Normal)	2 = Slave B Type CA Mode
					3 = Init as Master		3 = Slave B Type CA Mode

FQD			$6		0 = No Host Service Request	0 = Primary Channel  Normal Mode
Fast Quadrature				1 = Not Used			1 = Secondary Channel  Normal
Decode					2 = Read TCR1			    Mode
					3 = Initialize			2 = Primary Channel  Fast Mode
									3 = Secondary Channel  Fast Mode

CHANNEL CONTROL Options

TBS       PAC     PSC   Action
8 7 6 5   4 3 2   1 0   Input                     Output
                  0 0   -                         Force Pin as Specified by PAC Latches
                  0 1   -                         Force Pin High
                  1 0   -                         Force Pin Low
                  1 1   -                         Do Not Force Any State
          0 0 0   Do Not Detect Transition        Do Not Change Pin State on Match
          0 0 1   Detect Rising Edge              High on Match
          0 1 0   Detect Falling Edge             Low on Match
          0 1 1   Detect Either Edge              Toggle on Match
          1 x x   Do Not Change PAC               Do Not Change PAC
0 0 x x                  Input Channel            -
0 0 0 0                  Capture TCR1, Match TCR1 -
0 0 0 1                  Capture TCR1, Match TCR2 -
0 0 1 0                  Capture TCR2, Match TCR1 -
0 0 1 1                  Capture TCR2, Match TCR2 -
0 1 x x                  -                        Ouput Channel
0 1 0 0                  -                        Capture TCR1, Match TCR1
0 1 0 1                  -                        Capture TCR1, Match TCR2
0 1 1 0                  -                        Capture TCR2, Match TCR1
0 1 1 1                  -                        Capture TCR2, Match TCR2
1 x x x                  Do Not Change TBS        Do Not Change TBS

TBS  Time Base/Directionality Selection
    The TBS field specifies a channel pin as either input or output, and the time base to
    be used for match and capture events. Either TCR can be used for the match event or
    for the input capture event. For example, the match event of the channel may use one
    TCR as a time base, while the input capture event of the same channel may use the
    other TCR as a time base.
PAC  Pin Action Control
    The PAC field specifies the pin logic response as either a timer channel input or output.
    For input, PAC specifies the transition edge to be detected, resulting in assertion of
    the transition detect latch. For output, PAC specifies the logic level to be output to the
    pin due to assertion of the match recognition latch.
PSC  Pin State Control (PSC)
    The PSC field forces the output logic level of the pin directly, while not affecting the pin
    action control latches or the output level specified by the state of the pin action control
    latches.

************************************************************************* */

#include <linux/module.h>
#include <linux/version.h>
#include <linux/types.h>
#include <linux/errno.h>

#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/fcntl.h>
#include <linux/fs.h>


#if CONFIG_MODVERSIONS==1
#define MODVERSIONS
#include <linux/modversions.h>
#endif        


#include <linux/devfs_fs_kernel.h>


/*   #if LINUX_VERSION_CODE > KERNEL_VERSION(2,2,0)  */
#include <asm/uaccess.h> /* for put_user */
/*    #endif   */


#include <linux/poll.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <linux/init.h>   /* added by Ron so driver builds directly into kernel */


#include <asm/io.h>
#include <asm/irq.h>

#include <linux/signal.h>
#include <linux/sched.h>

#include <asm/m68332_exts.h>
// define IOCTL commands
#include <asm/tpu_ioctl.h>
#include "cpu32tpu.h"

#define DEVICE_NAME "tpu"


int __init tpu_init(void);
static   /* made static to avoid symbol clashes - SchS/Kanis GmbH 022005 */
int init(void);
void __exit tpu_exit(void);


/* struct wait_queue *wqueue;   */
static DECLARE_WAIT_QUEUE_HEAD(wqueue);   /* use ver 2.4 static declaration - ron */
  /* or should we use   wait_queue_heat_t *wqueue   ?? see page 141  */

/* preset values for all channels */
tpu_device tpu_dev [] =
	/*  fct sq sv pr ir st ov act setm  getm  clrm  own  wr rd  bufp */
	{ {   0, 0, 0, 0, 0, 1, 0, 0,    0,    0,    0, NULL, 0, 0, NULL },  /* channel 0 */
	  {   0, 0, 0, 0, 0, 1, 0, 0,    0,    0,    0, NULL, 0, 0, NULL },  /* channel 1 */
	  {   0, 0, 0, 0, 0, 1, 0, 0,    0,    0,    0, NULL, 0, 0, NULL },  /* channel 2 */
	  {   0, 0, 0, 0, 0, 1, 0, 0,    0,    0,    0, NULL, 0, 0, NULL },  /* channel 0 */
	  {   0, 0, 0, 0, 0, 1, 0, 0,    0,    0,    0, NULL, 0, 0, NULL },  /* channel 4 */
	  {   0, 0, 0, 0, 0, 1, 0, 0,    0,    0,    0, NULL, 0, 0, NULL },  /* channel 5 */
	  {   0, 0, 0, 0, 0, 1, 0, 0,    0,    0,    0, NULL, 0, 0, NULL },  /* channel 6 */
	  {   0, 0, 0, 0, 0, 1, 0, 0,    0,    0,    0, NULL, 0, 0, NULL },  /* channel 7 */
	  {   0, 0, 0, 0, 0, 1, 0, 0,    0,    0,    0, NULL, 0, 0, NULL },  /* channel 8 */
	  {   0, 0, 0, 0, 0, 1, 0, 0,    0,    0,    0, NULL, 0, 0, NULL },  /* channel 9 */
	  {   0, 0, 0, 0, 0, 1, 0, 0,    0,    0,    0, NULL, 0, 0, NULL },  /* channel 10 */
	  {   0, 0, 0, 0, 0, 1, 0, 0,    0,    0,    0, NULL, 0, 0, NULL },  /* channel 11 */
	  {   0, 0, 0, 0, 0, 1, 0, 0,    0,    0,    0, NULL, 0, 0, NULL },  /* channel 12 */
	  {   0, 0, 0, 0, 0, 1, 0, 0,    0,    0,    0, NULL, 0, 0, NULL },  /* channel 10 */
	  {   0, 0, 0, 0, 0, 1, 0, 0,    0,    0,    0, NULL, 0, 0, NULL },  /* channel 14 */
	  {   0, 0, 0, 0, 0, 1, 0, 0,    0,    0,    0, NULL, 0, 0, NULL },  /* channel 15 */
	};


void tpu_interrupt (int irq, void *dev_id, struct pt_regs *regs)
{
	int ch = irq- TPU_IRQ_VECTOR;
	tpu_device * dev = &(tpu_dev [ch]);
	unsigned short msk = dev->get_mask, clr = dev->clr_mask, n;
	PRAM_ch * params = &(Parameter_RAM [ch]);

	tpu.tpu_cisr &= ~(TPU_IR_MASK (ch));   /* clear channel interrupt request ??? */

	if (dev->dbufp) {
		n= 0;
		if (dev->overwrite_old) {
			while (msk) {
				if (msk & 1) {
					dev->dbufp [dev->written++]= (*params) [n];
					dev->written %= TPU_BUFSIZE;   /* wrap around */
					if (dev->written == dev->read) {  /* don't keep data not yet read */
						dev->read++;
						dev->read %= TPU_BUFSIZE;   /* wrap around */
					}
				}
				msk >>= 1; n++;
			}
		}
		else {
			while (msk && (dev->written != dev->read)) {  /* don't overwrite unread data */
				if (msk & 1) {
					dev->dbufp [dev->written++]= (*params) [n];
					dev->written %= TPU_BUFSIZE;   /* wrap around */
				}
				msk >>= 1; n++;
			}
		}
	}
	n= 0;
	while (clr) {
		if (clr & 1)
			(*params) [n]= 0;
		clr >>= 1; n++;
	}
	if (dev->callback)
		dev->callback (ch);
}


// static 
int tpu_ioctl (struct inode *inode, struct file *filep, unsigned int cmd, unsigned long arg)
{
        int minor = MINOR (filep->f_dentry->d_inode->i_rdev), rc = 0;
	tpu_device * dev = (tpu_device *) filep->private_data;
	tpu_device * data = (tpu_device *) arg;

	switch (cmd) {
		case TPUIOCS_SETPAR:		/* set channel function parameters */
			data->read=    dev->read;   /* keep buffer parameters... */
			data->written= dev->written;
			data->dbufp=   dev->dbufp;
			data->active=  dev->active;
//??			data->callback= dev->callback;  /* and call back address */
			*dev= *data;
			break;
		case TPUIOCS_GETPAR:		/* get channel function parameters */
			*data= *dev;
			data->dbufp= NULL;      /* do NOT pass data buffer! (not user accessible) */
			break;
		case TPUIOCS_INIT:		/* initialize channel, set data */
			if (dev->active) {
				tpu.tpu_cpr  [TPU_HST_OFF (minor)] &= ~(TPU_HST_MASK (minor));
				tpu.tpu_cfsr [TPU_FCT_OFF (minor)] |= dev->channel_func  << TPU_FCT_SHIFT (minor);
				tpu.tpu_hsqr [TPU_HST_OFF (minor)] |= dev->host_sequence << TPU_HST_SHIFT (minor);
				tpu.tpu_hsrr [TPU_HST_OFF (minor)] |= dev->host_service  << TPU_HST_SHIFT (minor);
				tpu.tpu_cier |= dev->interrupt << TPU_IR_SHIFT (minor);
				tpu.tpu_cpr  [TPU_HST_OFF (minor)] |= dev->priority      << TPU_HST_SHIFT (minor);
			}
			else
				rc= -EINVAL;  /* error if parameter RAM not yet set */
			break;
		case TPUIOCG_START:		/* start up channel (set priority) */
			if (tpu.tpu_cpr [TPU_HST_OFF (minor)] & TPU_HST_MASK (minor))  /* error if already set */
				rc= -EINVAL;
			else
				tpu.tpu_cpr [TPU_HST_OFF (minor)] |= (arg & 0x3) << TPU_HST_SHIFT (minor);
			break;
		case TPUIOCS_STOP:		/* stop channel (clear priority) */
			tpu.tpu_cpr [TPU_HST_OFF (minor)] &= ~(TPU_HST_MASK (minor));
			break;
		case TPUIOCS_SEQUENCE:		/* set host sequence bits */
			tpu.tpu_hsqr [TPU_HST_OFF (minor)] |= (arg & TPU_HST_MASK (minor)) << TPU_HST_SHIFT (minor);
			break;
		case TPUIOCS_SERVICE:		/* set host service bits */
//			if (tpu.tpu_hsrr [TPU_HST_OFF (minor)])  /* error if already set */
//				rc= -EINVAL;
//			else
				tpu.tpu_hsrr [TPU_HST_OFF (minor)] |= (arg & TPU_HST_MASK (minor)) << TPU_HST_SHIFT (minor);
			break;
		case TPUIOCS_CLSNOSTOP:		/* set continue operation on close */
			dev->stop_on_close= arg == 0;   /* 0 - stops, else will continue */
			break;
		case TPUIOCS_OVERWRITE:		/* set overwrite outdated data */
			dev->overwrite_old= arg != 0;   /* 0 - keep old data, else overwrite */
			break;
		case TPUIOCS_SETMCR:		/* set TPU master control register */
			tpu.tpu_tpumcr= arg;
		    /* Set up all the other registers here too, as this could probabely not be done earlier! (see below) */
			tpu.tpu_cier= 0;                /* disable channel interrupts */
			tpu.tpu_cpr [0]= 0;             /* and clear all */
			tpu.tpu_cpr [1]= 0;             /* channel priorities */
			tpu.tpu_ticr= CIRL (TPU_IRQ_LEVEL) | CIBV (TPU_IRQ_VECTOR);  /* tpu interrupt configuration */
			break;
		case TPUIOCS_GETACT:		/* get channel active status */
			*(int *) arg= dev->active;
			break;
		case TPUIOCS_SETCALL:		/* set channel callback function   - void arg (int) */
			(unsigned long) dev->callback= arg;   /* 0 - clears, else sets callback function */
			break;
		default:
			rc = -EINVAL;
			break;
	}
	return rc;
}


// static 
int tpu_open (struct inode *inode, struct file *file)
{
        int channel = MINOR (file->f_dentry->d_inode->i_rdev);
	unsigned short * buffer;

	if (channel >= TPU_CHANNELS)  /* ensure device is available */
		return -ENODEV;

/*    #if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
		MOD_INC_USE_COUNT;
   #endif   */

	file->private_data = &(tpu_dev [channel]);

	if (tpu_dev [channel].dbufp)                 /* if channel has data buffer already */
		return 0;

	if ((buffer= kmalloc (sizeof (unsigned short)* TPU_BUFSIZE, GFP_KERNEL)) == NULL) {
  /*  #if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
		MOD_DEC_USE_COUNT;
	#endif  */
		return -ENOMEM;
	}

	/* set default values */
	tpu_dev [channel].dbufp= buffer;
	tpu_dev [channel].active=  0;                /* Channel not working */
	tpu_dev [channel].written= 0;                /* Channel write data buffer */
	tpu_dev [channel].read=    TPU_BUFSIZE- 1;   /* Channel read data buffer -> empty */

	if (!(tpu.tpu_tpumcr & TPU_STOP))
		tpu.tpu_cpr [TPU_HST_OFF (channel)] &= ~(TPU_HST_MASK (channel));  /* clear channel priority */
	return 0;
}


static int tpu_release (struct inode *inode, struct file *file)
{
        int channel = MINOR (file->f_dentry->d_inode->i_rdev);
	tpu_device * dev = file->private_data;

	if (dev->stop_on_close) {
		tpu.tpu_cier &= ~(TPU_IR_MASK (channel));                          /* disable channel interrupt */
		tpu.tpu_cpr [TPU_HST_OFF (channel)] &= ~(TPU_HST_MASK (channel));  /* stop channel */
		kfree (dev->dbufp);                                                /* and return buffer */
		dev->dbufp= NULL;
	}
	dev->callback= NULL;                                                       /* ALWAYS clear call back routine */

/*  #if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	MOD_DEC_USE_COUNT;
	#endif    */
	
	return 0;
}


/*
   Strategy:
   The device minor number defines the tpu channel to use.

   After opening, the device is initialized, but not active.

   Writing to the device defines the parameter RAM data to set,
   only data according to the set_mask are used.
   The first write action activates the tpu channel.

   Reading the device returns data as defined by the get_mask
   and collected by interrupts executed since activation or
   the last read action. Reading data clears the data buffer.
   Writing does NOT clear the buffer.
*/
// static 
int tpu_write (struct file * filep, const char * buffer, size_t length, loff_t *off)
{
        int channel = MINOR (filep->f_dentry->d_inode->i_rdev), n;
	tpu_device * dev = filep->private_data;
	unsigned char msk = dev->set_mask;
	unsigned short * wbuf = (unsigned short *) buffer;

	if (msk) {
		for (n= 0; msk && (n < (length >> 1)); msk >>= 1, n++) {
			if (msk & 1) {
				get_user (Parameter_RAM [channel][n], wbuf);
				wbuf++;
			}
		}
	}
	if (!dev->active) {
		tpu.tpu_cfsr [TPU_FCT_OFF (channel)] |= dev->channel_func  << TPU_FCT_SHIFT (channel);
		tpu.tpu_hsqr [TPU_HST_OFF (channel)] |= dev->host_sequence << TPU_HST_SHIFT (channel);
		tpu.tpu_hsrr [TPU_HST_OFF (channel)] |= dev->host_service  << TPU_HST_SHIFT (channel);
		tpu.tpu_cier |= dev->interrupt << TPU_IR_SHIFT (channel);
		tpu.tpu_cpr  [TPU_HST_OFF (channel)] |= dev->priority      << TPU_HST_SHIFT (channel);
		dev->active= 1;
	}
	return n << 1;
}

//   basic 2.4 kernel function format
// ssize_t tpu_read(struct file* w12f, char * w12c, size_t w12d, loff_t * w12e) { ; }

// static 
ssize_t tpu_read (struct file * filep, char * buffer, size_t length, loff_t *off)

/******** older 2.0 kernel format **********
static int tpu_read(
        struct inode *inode,
        struct file *filep,
        char *buffer,
        int length)
********************************************/
{
	tpu_device * dev = filep->private_data;
	short n = (length >> 1), written = dev->written, toread = (dev->read+ 1) % TPU_BUFSIZE;
	unsigned char msk = dev->get_mask;

	if ((!msk) || (toread == written))
		return 0;   /* if no data requested or present return immediately */

	if (written > toread) {
		if ((written- toread) < n)    /* not enough data */
			n= written- toread;
		copy_to_user (buffer, &(dev->dbufp [toread]), 2* n);
		dev->read += n;
		return n* 2;
	}
	else if (written < toread) {          /* had wrap around */
		if ((TPU_BUFSIZE- toread) >= n) {  /* can copy directly */
			copy_to_user (buffer, &(dev->dbufp [toread]), length);
			dev->read += n;
			return length;
		}
		else {                             /* must split copy */
			copy_to_user (buffer, &(dev->dbufp [toread]), 2* (TPU_BUFSIZE- toread));
			toread= TPU_BUFSIZE- toread;
			n -= toread;
			if (written < n)      /* not enough data */
				n= written;
			copy_to_user (&(buffer [2* toread]), &(dev->dbufp [0]), 2* n);
			dev->read= n- 1;
			return (n+ toread)* 2;
		}
	}
	else                                       /* no data at all */
	    return 0;
} 


                 /*  fixed for 2.4 kernel, owner was ifdef'ed out for 2.0 kernel */
struct file_operations TFops = {
	owner:		THIS_MODULE,
	read:		tpu_read, 
	write:		tpu_write,
	ioctl:		tpu_ioctl,
	open:		tpu_open,
	release:	tpu_release  /* a.k.a. close */
};


static   /* made static to avoid symbol clashes - SchS/Kanis GmbH 022005 */
int
init ()
{
	int ch;
	/* common init: driver or module: */

	for (ch= 0; ch < TPU_CHANNELS; ch++)
		if (request_irq (TPU_IRQ_VECTOR+ ch, tpu_interrupt, SA_INTERRUPT, "cpu32 tpu", NULL)) {
			printk("tpu: Unable to attach tpu interrupt vector=%d\n", TPU_IRQ_VECTOR);
			for (; ch >= 0; ch--)
				free_irq (TPU_IRQ_VECTOR+ ch, NULL);
			return -EINVAL;
		}

/* The TPU MAY NOT be touched here if it should be settable
   at any other time in the life of the system !!!
   This may ONLY be done ONCE after reset!
   (There's an ioctl available for that purpose) */
//	tpu.tpu_tpumcr= TPU_MCR_INIT;   /* tpu module configuration */
/* And so NONE of the other TPU registers may be set either,
   as doing so without the TPU enabled will require activating the RESET line... */
//	tpu.tpu_cier= 0;                /* disable channel interrupts */
//	tpu.tpu_cpr [0]= 0;             /* and clear all */
//	tpu.tpu_cpr [1]= 0;             /* channel priorities */
//	tpu.tpu_ticr= CIRL (TPU_IRQ_LEVEL) | CIBV (TPU_IRQ_VECTOR);  /* tpu interrupt configuration */

	init_waitqueue_head (&wqueue);  /* was init_waitqueue() --Ron */
	printk ("68332 tpu driver ok\n");

	return 0;
}

/* init for compiled-in driver: call from mem.c */
int __init tpu_init ()                                  /* the __init added by ron  */
{
	int ret;

	if ((ret = devfs_register_chrdev (TPU_MAJOR, DEVICE_NAME, &TFops) < 0)) {
		printk ("%s device failed with %d\n",
			"Sorry, registering the character", ret);
		return ret;
	}
        printk ("tpu device driver installed OK\n");
	return init ();
}

/* Cleanup - undid whatever init_module did */
void __exit tpu_exit ()      /* the __exit added by ron  */
{
	int ret;

	tpu.tpu_cier= 0;             /* disable channel interrupts */
	tpu.tpu_cpr [0]= 0;          /* clear all */
	tpu.tpu_cpr [1]= 0;          /* channel priorities and */
	tpu.tpu_ticr= 0;             /* clear tpu interrupt configuration */
	tpu.tpu_tpumcr |= TPU_STOP;  /* halt tpu module */

	for (ret= 0; ret < TPU_CHANNELS; ret++)
		free_irq (TPU_IRQ_VECTOR+ ret, NULL);

	/* Unregister the device */
	if ((ret = devfs_unregister_chrdev (TPU_MAJOR, DEVICE_NAME)) < 0)
		printk("Error in unregister_chrdev: %d\n", ret);
}

module_init (tpu_init);      /*   added by ron so driver can compile directly into kernel */
module_exit (tpu_exit);      /*   added by ron so driver can compile directly into kernel */

MODULE_LICENSE ("GPL");
