/* cpu32tpu.h: Definitions for the mc68328 driver for the time processor unit interface module.
 *
 * Copyright (C) 2005       Sieghard Schicktanz <S.Schicktanz@Kanis.de>
 *                                                                      */
/* This program is free software; you can redistribute it and/or        */
/* modify it under the terms of the GNU Library General Public License  */
/* as published by the Free Software Foundation; either version 2       */
/* of the License, or (at your option) any later version.               */
/*                                                                      */
/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU Library General Public License for more details.                 */

#ifndef _CPU32_TPU_H
#define _CPU32_TPU_H

#include <asm/MC68332.h>
#include <asm/m68332_exts.h>

#include <linux/config.h>

#define TPU_MAJOR	125

#define TPU_CHANNELS	16


/* TPU register map
0xFFFE00    TPUMCR	TPU MODULE CONFIGURATION REGISTER
0xFFFE02    TCR		TPU TEST CONFIGURATION REGISTER
0xFFFE04    DSCR	DEVELOPMENT SUPPORT CONTROL REGISTER
0xFFFE06    DSSR	DEVELOPMENT SUPPORT STATUS REGISTER
0xFFFE08    TICR	TPU INTERRUPT CONFIGURATION REGISTER
0xFFFE0A    CIER	CHANNEL INTERRUPT ENABLE REGISTER
0xFFFE0C    CFSR0	CHANNEL FUNCTION SELECT REGISTER 0
0xFFFE0E    CFSR1	CHANNEL FUNCTION SELECT REGISTER 1
0xFFFE10    CFSR2	CHANNEL FUNCTION SELECT REGISTER 2
0xFFFE12    CFSR3	CHANNEL FUNCTION SELECT REGISTER 3

0xFFFE14    HSQR0	HOST SEQUENCE REGISTER 0
0xFFFE16    HSQR1	HOST SEQUENCE REGISTER 1
0xFFFE18    HSRR0	HOST SERVICE REQUEST REGISTER 0
0xFFFE1A    HSRR1	HOST SERVICE REQUEST REGISTER 1
0xFFFE1C    CPR0	CHANNEL PRIORITY REGISTER 0
0xFFFE1E    CPR1	CHANNEL PRIORITY REGISTER 1
0xFFFE20    CISR	CHANNEL INTERRUPT STATUS REGISTER
0xFFFE22    LR		LINK REGISTER
0xFFFE24    SGLR	SERVICE GRANT LATCH REGISTER
0xFFFE26    DCNR	DECODED CHANNEL NUMBER REGISTER
*/

/* TPU register bit definitions */

/* TPUMCR bits:
	STOP  Stop Bit
	TCR1P  Timer Count Register 1 Prescaler Control
	TCR2P  Timer Count Register 2 Prescaler Control
	EMU  Emulation Control
	T2CG  TCR2 Clock/Gate Control
	STF  Stop Flag
	SUPV  Supervisor/Unrestricted
	PSCK  Prescaler Clock
	IARB  Interrupt Arbitration
 */
#define TPU_STOP	0x8000
#define TPU_TCR1P_MASK	0x6000
#define TPU_TCR2P_MASK	0x1800
#define TPU_EMU		0x0400
#define TPU_T2CG	0x0200
#define TPU_STF		0x0100
#define TPU_SUPV	0x0080
#define TPU_PSCK	0x0040
#define TPU_IARB_MASK	0xf

/* TICR macros:
	CIRL  Channel Interrupt Request Level,
	CIBV  Channel Interrupt Base Vector
 */
#define CIRL(n)			((n) << 8)
#define CIBV(v)			((v) & 0xf0)

/* CFSR macros */
#define TPU_FCT_SHIFT(ch)	(((ch) & 3)* 4)
#define TPU_FCT_MASK(ch)	(0xf << (((ch) & 3)* 4))
#define TPU_FCT_OFF(ch)		((15- (ch)) >> 2)

/* HSQR / HSRR / CPR macros */
#define TPU_HST_SHIFT(ch)	(((ch) & 7)* 2)
#define TPU_HST_MASK(ch)	(3 << (((ch) & 7)* 2))
#define TPU_HST_OFF(ch)		((15- (ch)) >> 3)

/* CIER / CISR / LR / SGLR / DCNR macros */
#define TPU_IR_SHIFT(ch)	(ch)
#define TPU_IR_MASK(ch)		(1 << (ch))

/* CFSR0 bits */
#define TPU_FCT_CH15	0xf000
#define TPU_FCT_CH14	0x0f00
#define TPU_FCT_CH13	0x00f0
#define TPU_FCT_CH12	0x000f

/* CFSR1 bits */
#define TPU_FCT_CH11	0xf000
#define TPU_FCT_CH10	0x0f00
#define TPU_FCT_CH9	0x00f0
#define TPU_FCT_CH8	0x000f

/* CFSR2 bits */
#define TPU_FCT_CH7	0xf000
#define TPU_FCT_CH6	0x0f00
#define TPU_FCT_CH5	0x00f0
#define TPU_FCT_CH4	0x000f

/* CFSR3 bits */
#define TPU_FCT_CH3	0xf000
#define TPU_FCT_CH2	0x0f00
#define TPU_FCT_CH1	0x00f0
#define TPU_FCT_CH0	0x000f

/* HSQR0 / HSRR0 / CPR0 bits */
#define TPU_H_CH15	0xc000
#define TPU_H_CH14	0x3000
#define TPU_H_CH13	0x0c00
#define TPU_H_CH12	0x0300
#define TPU_H_CH11	0x00c0
#define TPU_H_CH10	0x0030
#define TPU_H_CH9	0x000c
#define TPU_H_CH8	0x0003

/* HSQR1 / HSRR1 / CPR1 bits */
#define TPU_H_CH7	0xc000
#define TPU_H_CH6	0x3000
#define TPU_H_CH5	0x0c00
#define TPU_H_CH4	0x0300
#define TPU_H_CH3	0x00c0
#define TPU_H_CH2	0x0030
#define TPU_H_CH1	0x000c
#define TPU_H_CH0	0x0003

/* CIER / CISR / LR / SGLR / DCNR bits */
#define TPU_B_CH15	0x8000
#define TPU_B_CH14	0x4000
#define TPU_B_CH13	0x2000
#define TPU_B_CH12	0x1000
#define TPU_B_CH11	0x0800
#define TPU_B_CH10	0x0400
#define TPU_B_CH9	0x0200
#define TPU_B_CH8	0x0100
#define TPU_B_CH7	0x0080
#define TPU_B_CH6	0x0040
#define TPU_B_CH5	0x0020
#define TPU_B_CH4	0x0010
#define TPU_B_CH3	0x0008
#define TPU_B_CH2	0x0004
#define TPU_B_CH1	0x0002
#define TPU_B_CH0	0x0001

/* TPU Function definitions */
/* Mask Set A Time Function Encodings */
#define TPU_A_PPWA	0xF
#define TPU_A_OC	0xE
#define TPU_A_SM	0xD
#define TPU_A_PSP	0xC
#define TPU_A_PMAPMM	0xB
#define TPU_A_ITC	0xA
#define TPU_A_PWM	0x9
#define TPU_A_DIO	0x8
#define TPU_A_SPWM	0x7
#define TPU_A_QDEC	0x6

/* Mask Set G Time Function Encodings */
#define TPU_G_PTA	0xF
#define TPU_G_QO	0xE
#define TPU_G_TSM	0xD
#define TPU_G_FQM	0xC
#define TPU_G_UART	0xB
#define TPU_G_NITC	0xA
#define TPU_G_COMM	0x9
#define TPU_G_HALLD	0x8
#define TPU_G_MCPWM	0x7
#define TPU_G_FQD	0x6


#define Parameter_RAM	((PRAM_ch*) 0xFFFFFF00)
// #define Parameter_RAM	int (*var)[8]

#ifdef CONFIG_VTSPLUS
// define TPU base setup, all 0 except arbitration level
 #define TPU_MCR_INIT	13

 #define TPU_IRQ_LEVEL	5
// first of sequence of 16 vectors, only high nibble settable!
 #define TPU_IRQ_VECTOR	64

 #define TPU_BUFSIZE	512
#endif


#endif /* !(_CPU32_TPU_H) */
